from contextlib import suppress
from json import JSONDecodeError, loads
from pathlib import Path

from ccflow import CallableModel, ContextBase, Flow, ResultBase
from pydantic import Field, PrivateAttr, model_validator

from .config import Configuration, Outputs
from .parameters import PapermillParameters, Parameters

__all__ = ("Executor",)


class ExecutorParameters(ContextBase):
    parameters: list[Parameters]


class ExecutorOutputs(ResultBase):
    outputs: list[Outputs]


class Executor(CallableModel):
    # Common with Configuration
    debug: bool | None = Field(default=False)

    nbprint: Configuration
    parameters: list[Parameters] = Field(default_factory=list)
    outputs: list[Outputs] = Field(default_factory=list)

    _nbprints: list[Configuration] = PrivateAttr(default_factory=list)
    _parameters: list[Parameters] = PrivateAttr(default_factory=list)

    @model_validator(mode="before")
    @classmethod
    def _parameters_list_to_list(cls, values) -> dict:
        if "parameters" in values:
            if isinstance(values["parameters"], str):
                # If its a file, try to read it
                parameters_str = values["parameters"]
                if Path(parameters_str).exists():
                    if parameters_str.endswith((".json",)):
                        with open(parameters_str) as f:
                            values["parameters"] = f.read()
                    elif parameters_str.endswith((".jsonl", ".yml")):
                        with open(parameters_str) as f:
                            values["parameters"] = f"[{f.read()}]"
                with suppress(JSONDecodeError):
                    # Try parse as json
                    values["parameters"] = loads(values["parameters"])
            if isinstance(values["parameters"], dict):
                values["parameters"] = [PapermillParameters(**values["parameters"])]
            if isinstance(values["parameters"], list):
                values["parameters"] = [p if isinstance(p, Parameters) else PapermillParameters(**p) for p in values["parameters"]]
        return values

    @Flow.deps
    def __deps__(self, context: ExecutorParameters) -> list[tuple[Configuration, PapermillParameters]]:
        # TODO: Make sure outputs will not clobber
        # TODO: vars is ugly hack, need to fix in papermill parameters
        self._nbprints = [self.nbprint.model_copy(deep=True, update={"_multi": True}) for _ in context.parameters]
        self._parameters = context.parameters
        return list(
            zip(
                self._nbprints,
                [[_] for _ in self._parameters],
                strict=True,
            )
        )

    @Flow.call
    def __call__(self, context: ExecutorParameters = None) -> ExecutorOutputs:  # noqa: ARG002
        op = []
        for p in self._parameters:
            nb = self.nbprint.model_copy(deep=True, update={"_multi": True})
            op.append(nb(p.model_copy(deep=True)))
        self.outputs = op
        return ExecutorOutputs(outputs=op)

    def run(self, dry_run: bool = False) -> ExecutorOutputs:
        if dry_run:
            raise NotImplementedError
        res = self(ExecutorParameters(parameters=self.parameters))
        # Run outputs postprocessing
        if res.outputs and res.outputs[0].postprocess:
            # Run postprocessing
            res.outputs[0].postprocess.object(self._nbprints)
        return res



from IPython.display import HTML, display
from pydantic import Field, field_validator

from .base import Content


class ContentPageBreak(Content):
    tags: list[str] = Field(default_factory=list)

    def __call__(self, **_) -> None:
        display(HTML('<p class="pagebreak"></p>'))

    @field_validator("tags", mode="after")
    @classmethod
    def _ensure_tags(cls, v: list[str]) -> list[str]:
        if "nbprint:content:pagebreak" not in v:
            v.append("nbprint:content:pagebreak")
        return v



import os
from pathlib import Path
from typing import Literal

from ccflow import PyObjectPath
from nbformat import NotebookNode
from pydantic import Field, PrivateAttr, field_validator

from nbprint.config import Configuration, Outputs, OutputsProcessing

__all__ = ("HTMLOutputs", "NBConvertOutputs", "NBConvertShortCircuitOutputs", "NotebookOutputs", "PDFOutputs", "WebHTMLOutputs", "short_circuit_hook")


class NBConvertOutputs(Outputs):
    target: Literal["ipynb", "notebook", "html", "webhtml", "pdf", "webpdf"] | None = "html"  # TODO: nbconvert types
    execute: bool | None = True
    timeout: int | None = 600
    template: str | None = "nbprint"

    # TODO: maybe allow collecting by index
    # collect_cells: list[int] = Field(default=[], description="List of cell indices to collect outputs from.")
    collect_outputs: bool = Field(
        default=False, description=("Whether to collect cell outputs into the context. Cells with tag `nbprint:output:<key>` will be collected under `<key>`.")
    )
    execute_hook: PyObjectPath | None = Field(
        default=None,
        description=(
            "A callable hook that is called after nbconvert execution of the notebook. "
            "It is passed the config instance. "
            "If it returns something non-None, that value is returned by `run` instead of the output path."
            "NOTE: Parent/child class hooks may also be called."
        ),
    )
    nbconvert_hook: PyObjectPath | None = Field(
        default=None,
        description=(
            "A callable hook that is called after nbconvert of the previously executed notebook. "
            "It is passed the config instance. "
            "If it returns something non-None, that value is returned by `run` instead of the output path."
            "NOTE: Parent/child class hooks may also be called."
        ),
    )

    _collected_cells: dict[int | str, list[dict[str, str]]] = PrivateAttr(default_factory=dict)

    @property
    def outputs(self) -> dict[int | str, list[dict[str, str]]]:
        # NOTE: parent class has `output`
        return self._collected_cells

    @field_validator("target", mode="before")
    @classmethod
    def validate_target(cls, v) -> str:
        if v is None:
            return "html"
        if v == "pdf":
            return "webpdf"
        if v == "notebook":
            return "ipynb"
        return v

    def _compute_outputs(self, config: "Configuration") -> None:
        super()._compute_outputs(config=config)
        # Update executed path if needed
        if self.execute:
            self._nb_executed_path = self.notebook.parent / f"{self.notebook.stem}.executed.ipynb"
        # Update output path
        if self.target == "webpdf":
            target = "pdf"
        elif self.target == "webhtml":
            target = "html"
        else:
            target = self.target
        if self.target == "ipynb" and self.execute:
            self._output_path = Path(str(self.output).replace(".ipynb", ".executed.ipynb"))
        else:
            self._output_path = Path(str(self.output).replace(".ipynb", f".{target}"))

    @staticmethod
    def _get_output_key(cell: NotebookNode) -> str | None:
        """Get the output key from cell metadata or tags."""
        if "nbprint" in cell.metadata and "output" in cell.metadata.nbprint:
            return cell.metadata.nbprint.output
        for tag in cell.metadata.get("tags", []):
            if tag.startswith("nbprint:output:"):
                return tag.split("nbprint:output:")[1]
        return None

    def _extract_cell_outputs(self) -> None:
        """Extract outputs from selected cells into the context."""
        # We're going to:
        # - read the notebook
        # - go through each cell and look for nbprint metadata
        #   - either `nbprint:output:<key>` tag or
        #   - `nbprint` metadata with `output` key
        # - collect outputs from those cells into self._collected_cells, such that:
        #     - the mimetype is used to determine the type of output
        #     - if we know how to deal, store natively
        #     - else, store as-is

        from nbformat import reads

        notebook_content = self.executed_notebook.read_text()
        nb = reads(notebook_content, as_version=4)

        for cell in nb.cells:
            if "nbprint" not in cell.metadata and not any(tag.startswith("nbprint:output:") for tag in cell.metadata.get("tags", [])):
                continue

            output_key = self._get_output_key(cell)
            if output_key is None:
                continue

            outputs = []
            for output in cell.get("outputs", []):
                output_data = {}
                if "data" in output:
                    output_data = dict(output["data"].items())
                elif "text" in output:
                    output_data["text/plain"] = output["text"]
                outputs.append(output_data)
            if output_key not in self._collected_cells:
                self._collected_cells[output_key] = []
            self._collected_cells[output_key].extend(outputs)

    def run(self, config: "Configuration", gen: NotebookNode) -> Path:
        from nbconvert.nbconvertapp import main as execute_nbconvert

        # Run parent to create notebook
        notebook = super().run(config=config, gen=gen)

        # If notebook is None, we stop
        if notebook in (None, OutputsProcessing.STOP):
            return OutputsProcessing.STOP

        # TODO: fix in nbconvert
        output = str(self.output).replace(".webpdf", ".pdf").replace(".pdf", "") if self.target == "webpdf" else str(self.output)

        cmd = [
            str(notebook),
            f"--to={self.target}",
            f"--output={output}",
            f"--template={self.template}",
        ]

        # We have some cheats here because we have to
        os.environ["_NBPRINT_IN_NBCONVERT"] = "1"
        os.environ["PSP_JUPYTER_HTML_EXPORT"] = "1"

        if self.execute:
            nbex_cmd = [
                str(notebook),
                "--to=notebook",
                f"--output={self.executed_notebook!s}",
                "--execute",
                f"--ExecutePreprocessor.timeout={self.timeout}",
            ]

            # Update cmd to use executed notebook
            cmd[0] = str(self.executed_notebook)

            # Execute nbconvert
            execute_nbconvert(nbex_cmd)

            # Extract cells by tags
            self._extract_cell_outputs()

            if self.execute_hook and self.execute_hook.object(config) in (OutputsProcessing.STOP, None):
                return OutputsProcessing.STOP

        if not (self.execute and self.target == "ipynb"):
            # If target is notebook, we already did it above
            execute_nbconvert(cmd)

        if self.nbconvert_hook and self.nbconvert_hook.object(config) in (OutputsProcessing.STOP, None):
            return OutputsProcessing.STOP
        return self.output


class NotebookOutputs(NBConvertOutputs):
    target: Literal["ipynb"] = "ipynb"


class HTMLOutputs(NBConvertOutputs):
    target: Literal["html"] = "html"


class WebHTMLOutputs(NBConvertOutputs):
    target: Literal["webhtml"] = "webhtml"


class PDFOutputs(NBConvertOutputs):
    target: Literal["webpdf"] = "webpdf"


def short_circuit_hook(config: "Configuration") -> OutputsProcessing | bool:
    """A hook that short-circuits processing if a certain cell returns True."""
    return (
        OutputsProcessing.STOP
        if config.outputs.outputs
        and "stop" in config.outputs.outputs
        and any(outcome.get("text/plain", "").strip().lower() == "true" for outcome in config.outputs.outputs["stop"])
        else True
    )


class NBConvertShortCircuitOutputs(NBConvertOutputs):
    """A specialized NBConvertOutputs that installs a default hook to stop processing if a certain cell
    with tag nbprint:output:stop returns True.
    """

    execute_hook: PyObjectPath = Field(
        default=PyObjectPath("nbprint.config.outputs.nbconvert.short_circuit_hook"),
        description="A hook that short-circuits processing if a certain cell with tag nbprint:output:stop returns True.",
    )



import ast
from json import dumps
from typing import Any, Dict

from ccflow import ContextBase
from nbformat import NotebookNode
from pydantic import Field, field_validator, model_serializer, model_validator

from nbprint.config.base import BaseModel, Role

__all__ = ("PapermillParameters", "Parameters")


class Parameters(ContextBase, BaseModel):
    tags: list[str] = Field(default_factory=list)
    role: Role = Role.PARAMETERS
    ignore: bool = True

    @field_validator("tags", mode="after")
    @classmethod
    def _ensure_tags(cls, v: list[str]) -> list[str]:
        if "nbprint:parameters" not in v:
            v.append("nbprint:parameters")
        if "parameters" not in v:
            v.append("parameters")
        return v

    def generate(self, metadata: dict, **_) -> NotebookNode:
        cell = self._base_generate_meta(metadata=metadata)
        # if nb_vars:
        #     # add parameter variable
        #     nb_vars.add(k)

        mod = ast.Module(body=[], type_ignores=[])

        # NOTE: use model_dump(mode="json") here to be compatible with
        # papermill-based json parameters
        for i, (k, v) in enumerate(self.model_dump(mode="json", exclude={"type", "tags", "role", "ignore"}).items()):
            if isinstance(v, bool):
                # Handle separately
                to_write = str(v)
            else:
                to_write = dumps(v)
                if isinstance(to_write, str) and (": true," in to_write or ": false," in to_write):
                    to_write = to_write.replace(": true,", ": True,").replace(": false,", ": False,")
            mod.body.append(
                ast.Assign(
                    targets=[ast.Name(id=k, ctx=ast.Store())],
                    value=ast.parse(to_write, mode="eval").body,
                    lineno=i,
                )
            )
        source = ast.unparse(mod).replace('"', '\\"')
        cell.source = source
        return cell


class PapermillParameters(Parameters):
    """Papermill parameters function implicitly as a dict"""

    vars: Dict[str, Any] = Field(default_factory=dict)

    def generate(self, metadata: dict, **_) -> NotebookNode:
        cell = self._base_generate_meta(metadata=metadata)
        mod = ast.Module(body=[], type_ignores=[])

        # Create a dictionary assignment for parameters
        for i, (k, v) in enumerate(self.vars.items()):
            if isinstance(v, bool):
                to_write = str(v)
            else:
                to_write = dumps(v)
                if isinstance(to_write, str) and (": true," in to_write or ": false," in to_write):
                    to_write = to_write.replace(": true,", ": True,").replace(": false,", ": False,")
            mod.body.append(
                ast.Assign(
                    targets=[ast.Name(id=k, ctx=ast.Store())],
                    value=ast.parse(to_write, mode="eval").body,
                    lineno=i,
                )
            )
        source = ast.unparse(mod).replace('"', '\\"')
        cell.source = source
        return cell

    @model_validator(mode="before")
    @classmethod
    def _model_before_validator(cls, data) -> dict:
        # Move all fields except those defined in Parameters to vars
        params_fields = set(cls.model_fields.keys()) | {"_target_"}
        vars_dict = {k: v for k, v in data.items() if k not in params_fields}
        for k in vars_dict:
            data.pop(k)
        if "vars" not in data:
            data["vars"] = {}
        data["vars"].update(vars_dict)
        return data

    @model_serializer(mode="wrap")
    def _serialize_model(self, handler) -> dict[str, object]:
        serialized = handler(self)
        serialized = {k: v for k, v in serialized.items() if k != "vars"}
        serialized.update(self.vars)
        return serialized

    # NOTE: this shouldve been possible via a wrap or before validator,
    # but alas i could not get it to work
    def __setattr__(self, name: str, value) -> None:
        if name in self.__class__.model_fields and name != "vars":
            super().__setattr__(name, value)
        elif name == "vars":
            self.vars.update(value)
        else:
            self.vars[name] = value



from ast import literal_eval
from logging import getLogger
from pathlib import Path
from typing import List, Optional

from ccflow import ModelRegistry
from hydra import compose, initialize_config_dir
from nbformat import read as nb_read
from omegaconf import DictConfig, OmegaConf, open_dict

__all__ = ("load_config",)

_logger = getLogger(__name__)


def load_config(
    path: str,
    overrides: List[str] | None = None,
) -> dict:
    # convert to Path
    path = Path(path)

    if not isinstance(overrides, list):
        # maybe running via python, reset
        overrides = []

    # prune any empty strings
    overrides = [o for o in overrides if o]

    # TODO: right now, nbprint runs off a specific config file, whereas
    # hydra takes a config dir and config name. For now we use the nbprint
    # style, so we adjust accordingly
    if path.suffix in (".yaml",):
        config_dir = str(path.parent.absolute().resolve())
        config_name = str(path.name)
        is_default = False
    else:
        # Use base
        config_dir = str((Path(__file__).parent).absolute().resolve())
        config_name = "base.yaml"
        is_default = True

    with initialize_config_dir(config_dir=config_dir, version_base=None):
        if not is_default:
            cfg = compose(config_name=config_name, overrides=[], return_hydra_config=True)
            searchpaths = cfg["hydra"]["searchpath"]
            searchpaths.extend([config_dir])
            overrides = [*overrides.copy(), f"hydra.searchpath=[{','.join(searchpaths)}]"]
        cfg = compose(config_name=config_name, overrides=overrides)

        if "nbprint" not in cfg:
            _logger.warning("No 'nbprint' config found in the provided configuration. Assuming entire config is for nbprint.")
            with open_dict(cfg):
                if "_target_" not in cfg:
                    cfg._target_ = "nbprint.Configuration"
                cfg.nbprint = cfg.copy()

        # bridge hydra and non-hydra
        if "name" not in cfg.nbprint:
            with open_dict(cfg.nbprint):
                cfg.nbprint.name = path.name.replace(".yaml", "").replace(".ipynb", "")

        # If its a notebook, parse it out and run directly
        # Read notebook contents and shove into config
        if path.suffix in (".ipynb",):
            with open_dict(cfg):
                cfg.nbprint.notebook = path

        if isinstance(cfg, DictConfig):
            cfg = OmegaConf.to_container(cfg, resolve=True)

        registry = ModelRegistry.root()
        registry.load_config(cfg=cfg, overwrite=True)

        if "callable" in cfg:
            registry.add("callable", registry[cfg["callable"]], overwrite=True)
        elif "callable" in cfg.get("nbprint", {}):
            registry.add("callable", registry[cfg["nbprint"]["callable"]], overwrite=True)
        else:
            registry.add("callable", registry["nbprint"], overwrite=True)

    return registry



from typing import Any, Dict

import pandas as pd
from pydantic import Field

from nbprint import Context

__all__ = ("ExampleContext",)


class ExampleContext(Context):
    string: str = ""
    df: pd.DataFrame | None = None
    params: Dict[str, Any] = Field(default_factory=dict)



from datetime import date, datetime
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING
from uuid import uuid4

from ccflow import PyObjectPath, ResultBase
from IPython.display import HTML
from jinja2 import Template
from nbformat import NotebookNode, writes
from pydantic import Field, PrivateAttr, field_validator, model_validator

from nbprint.config.base import BaseModel, Role

if TYPE_CHECKING:
    from .config import Configuration

__all__ = ("Outputs", "OutputsProcessing")


class OutputsProcessing(str, Enum):
    CONTINUE = "continue"
    STOP = "stop"


class Outputs(ResultBase, BaseModel):
    root: Path = Field(default=Path.cwd() / "outputs")
    naming: str = Field(default="{{name}}-{{date}}")

    tags: list[str] = Field(default_factory=list)
    role: Role = Role.OUTPUTS
    ignore: bool = True

    embedded: bool = Field(default=False, description="Whether this output is expected to run from its embedding inside the notebook.")
    hook: PyObjectPath | None = Field(
        default=None,
        description=(
            "A callable hook that is called after generation of the notebook. "
            "It is passed the config instance. "
            "If it returns something non-None, that value is returned by `run` instead of the output path."
        ),
    )
    postprocess: PyObjectPath | None = Field(
        default=None,
        description=(
            "A callable hook that is called after all processing completes. "
            "It is passed the config instance/s. "
            "NOTE: It may receive multiple Configuration instances, if a parameterized run was performed."
        ),
    )

    # Private variables to compute and store paths
    _nb_path: Path | None = PrivateAttr(default=None)
    _nb_executed_path: Path | None = PrivateAttr(default=None)
    _output_path: Path | None = PrivateAttr(default=None)

    @property
    def notebook(self) -> Path:
        return self._nb_path

    @property
    def executed_notebook(self) -> Path:
        return self._nb_executed_path

    @property
    def output(self) -> Path:
        return self._output_path

    @field_validator("tags", mode="after")
    @classmethod
    def _ensure_tags(cls, v: list[str]) -> list[str]:
        if "nbprint:outputs" not in v:
            v.append("nbprint:outputs")
        return v

    @field_validator("root", mode="before")
    @classmethod
    def _convert_str_to_path(cls, v) -> Path:
        if isinstance(v, str):
            v = Path(v)
        if isinstance(v, Path):
            return v
        raise TypeError

    @model_validator(mode="after")
    def _ensure_embedded_or_cells_not_both(self) -> "Outputs":
        if self.embedded and self.cells:
            err = "Cannot set both 'embedded' and 'cells'. If `embedded`, set fields on context object. Otherwise, select cells."
            raise ValueError(err)
        return self

    def _compute_outputs(self, config: "Configuration") -> None:
        name = Template(self.naming).render(
            name=self._get_name(config=config),
            date=self._get_date(config=config),
            datetime=self._get_datetime(config=config),
            uuid=self._get_uuid(config=config),
            sha=self._get_sha(config=config),
            **config.parameters.model_dump(exclude_none=True, exclude_unset=True),
        )

        root = Path(self.root).resolve()
        self._nb_executed_path = root / f"{name}.ipynb"
        self._nb_path = root / f"{name}.ipynb"
        self._output_path = root / f"{name}.ipynb"

    def _get_name(self, config: "Configuration") -> str:
        return config.name

    def _get_date(self, **_) -> str:
        return date.today().isoformat()

    def _get_datetime(self, **_) -> str:
        return datetime.now().isoformat()

    def _get_uuid(self, **_) -> str:
        return uuid4()

    def _get_sha(self, config: "Configuration") -> str:
        import hashlib

        return hashlib.sha256(config.model_dump_json(by_alias=True).encode()).hexdigest()

    def run(self, config: "Configuration", gen: NotebookNode) -> Path:
        # Create file or folder path
        if not self.notebook.parent.exists():
            self.notebook.parent.mkdir(parents=True, exist_ok=True)

        self.notebook.write_text(writes(gen), encoding="utf-8")

        if self.hook and self.hook.object(config) in (OutputsProcessing.STOP, None):
            return OutputsProcessing.STOP
        return self.notebook

    def generate(self, metadata: dict, **kwargs) -> NotebookNode:
        return super().generate(metadata=metadata, **kwargs)

    def __call__(self, **_) -> HTML:
        if self.embedded:
            # If we're running this code, the notebook is already executing.
            # For child classes, this might mean saving ourselves (the notebook
            # in which we're running), then running nbconvert on ourselves
            # with execute=False.
            #
            # For now, this is a TODO.
            err = "Outputs cell execution in context is not yet implemented."
            raise NotImplementedError(err)
        return HTML("")



import os

from IPython.display import HTML

from nbprint import Content

__all__ = ("PlotlyRendererConfiguration",)


class PlotlyRendererConfiguration(Content):
    default: str | None = "notebook_connected"

    def __call__(self, **_) -> HTML:
        if os.environ.get("_NBPRINT_IN_NBCONVERT", "0") == "1":
            import plotly.io as pio

            pio.renderers.default = self.default
        return HTML("")



from ast import literal_eval
from pathlib import Path
from pprint import pprint
from sys import version_info
from typing import Type, Union

from ccflow import CallableModel, ContextType, Flow, ResultType
from hydra import compose, initialize_config_dir
from hydra.utils import instantiate
from nbformat import NotebookNode, read as nb_read
from nbformat.v4 import new_notebook
from pkn import getSimpleLogger
from pydantic import Field, PrivateAttr, field_validator, model_validator
from typing_extensions import Self

from nbprint import __version__
from nbprint.config.base import BaseModel, Role, _append_or_extend
from nbprint.config.content import Content, ContentCode, ContentMarkdown
from nbprint.config.exceptions import NBPrintPathIsYamlError, NBPrintPathOrModelMalformedError
from nbprint.config.page import Page

from .content import ContentMarshall
from .context import Context
from .outputs import Outputs, OutputsProcessing
from .parameters import PapermillParameters, Parameters

__all__ = (
    "Configuration",
    "load",
)

_log = getSimpleLogger("nbprint.config.core.config")


class Configuration(CallableModel, BaseModel):
    name: str
    resources: dict[str, BaseModel] = Field(default_factory=dict)
    outputs: Outputs
    parameters: Parameters = Field(default_factory=Parameters)
    page: Page = Field(default_factory=Page)
    context: Context = Field(default_factory=Context)

    content: ContentMarshall = Field(default_factory=ContentMarshall)

    # basic metadata
    tags: list[str] = Field(default_factory=list)
    role: Role = Role.CONFIGURATION
    ignore: bool = True
    pagedjs: bool = True
    debug: bool = True

    # internals
    _multi: bool = PrivateAttr(default=False)
    _nb_var_name: str = PrivateAttr(default="nbprint_config")
    _nb_vars: set = PrivateAttr(default_factory=set)

    @field_validator("tags", mode="after")
    @classmethod
    def _ensure_tags(cls, v: list[str]) -> list[str]:
        if "nbprint:config" not in v:
            v.append("nbprint:config")
        return v

    @field_validator("resources", mode="before")
    @classmethod
    def _convert_resources_from_obj(cls, value) -> dict[str, BaseModel]:
        if value is None:
            value = {}
        if isinstance(value, dict):
            for k, v in value.items():
                value[k] = BaseModel._to_type(v)
        return value

    @field_validator("outputs", mode="before")
    @classmethod
    def _convert_outputs_from_obj(cls, v) -> Outputs:
        return BaseModel._to_type(v, Outputs)

    @field_validator("parameters", mode="before")
    @classmethod
    def _convert_parameters_from_obj(cls, v) -> Parameters:
        return BaseModel._to_type(v, Parameters)

    @field_validator("page", mode="before")
    @classmethod
    def _convert_page_from_obj(cls, v) -> Page:
        return BaseModel._to_type(v, Page)

    @field_validator("context", mode="before")
    @classmethod
    def _convert_context_from_obj(cls, v) -> Context:
        return BaseModel._to_type(v, Context)

    @staticmethod
    def _convert_content_from_list(v) -> ContentMarshall:
        for i, element in enumerate(v):
            if isinstance(element, str):
                v[i] = Content(type_=element)
            elif isinstance(element, dict):
                v[i] = BaseModel._to_type(element)
        return ContentMarshall(middlematter=v)

    @staticmethod
    def _convert_content_from_dict(v) -> ContentMarshall:
        for key in ContentMarshall.model_fields:
            if key in v and isinstance(v[key], list):
                v[key] = Configuration._convert_content_from_list(v[key]).all
        return ContentMarshall(**v)

    @field_validator("content", mode="before")
    @classmethod
    def convert_content_from_obj(cls, v) -> ContentMarshall:
        if v is None:
            return ContentMarshall()
        if isinstance(v, list):
            return cls._convert_content_from_list(v)
        if isinstance(v, dict):
            return cls._convert_content_from_dict(v)
        return v

    @model_validator(mode="after")
    def _validate(self) -> Self:
        self.context.parameters = self.parameters
        self.outputs._compute_outputs(config=self)
        return self

    # NOTE: this shouldve been possible via a wrap validator,
    # but alas i could not get it to work
    def __setattr__(self, name: str, value) -> None:
        if name == "parameters" and self.parameters:
            value = BaseModel._to_type(value, Parameters) if isinstance(value, dict) and "type_" in value else PapermillParameters.model_validate(value)

            if not type(value) is not type(self.parameters):
                # replace wholesale
                super().__setattr__(name, value)
            else:
                # Union new parameters with existing parameters
                for k, v in value.model_dump(mode="json").items() if isinstance(value, Parameters) else value.items():
                    setattr(self.parameters, k, v)
            self._validate()
        return super().__setattr__(name, value)

    @staticmethod
    def _init_content(values) -> ContentMarshall:
        if "content" not in values:
            values["content"] = ContentMarshall()
        elif isinstance(values["content"], list):
            values["content"] = ContentMarshall(middlematter=values["content"])
        elif isinstance(values["content"], dict):
            values["content"] = ContentMarshall(**values["content"])
        else:
            e = f"Unexpected content format when loading from notebook: {type(values['content'])}"
            raise RuntimeError(e)

    @staticmethod
    def _cell_to_content(cell) -> Content:
        source = cell.source.strip()
        if not source:
            # skip empty cells
            return None
        # Cells may have nbprint metadata from the UI extension
        # "nbprint": {
        #     "attrs": "",
        #     "class": "",
        #     "class_selector": "",
        #     "css": "",
        #     "data": "{}",
        #     "element_selector": "",
        #     "esm": "",
        #     "id": "",
        #     "ignore": true,
        #     "role": "parameters",
        #     "type_": "nbprint.config.core.parameters.Parameters"
        #    },
        if "metadata" in cell and "nbprint" in cell["metadata"]:
            nbprint_cell_meta = cell["metadata"]["nbprint"]

            # TODO: not all fields are serdes symmetric
            nbprint_cell_meta.pop("attrs", None)
            nbprint_cell_meta.pop("class", None)
            nbprint_cell_meta.pop("class_selector", None)
            nbprint_cell_meta.pop("element_selector", None)
        else:
            nbprint_cell_meta = {}

        # Attach cell tags in to content
        if "tags" in cell["metadata"]:
            nbprint_cell_meta["tags"] = cell["metadata"]["tags"]

        # If this is an nbprint defined type, use that
        if "type_" in nbprint_cell_meta:
            content_type = nbprint_cell_meta["type_"]
            content_model = BaseModel._to_type({"type_": content_type}, Content)
            return content_model.model_validate(nbprint_cell_meta)

        # Set source
        nbprint_cell_meta["content"] = cell.source

        # Default handling: treat as code or markdown content
        if cell.cell_type in {"code"}:
            content = ContentCode.model_validate(nbprint_cell_meta)
        elif cell.cell_type in {"markdown"}:
            content = ContentMarkdown.model_validate(nbprint_cell_meta)
        else:
            # Skip, log warning
            _log.warning(f"Unsupported cell type when loading from notebook: {cell.cell_type}")
        return content

    @staticmethod
    def _parse_parameters_cell(cell) -> dict:
        new_parameters = {}
        param_lines = cell.source.splitlines()
        for line in param_lines:
            if "=" in line:
                key, value = line.split("=", 1)
                key = key.strip()
                value = value.strip()
                # Attempt to eval the value to get correct type
                try:
                    evaluated_value = literal_eval(value)
                except SyntaxError:
                    evaluated_value = value

                new_parameters[key] = evaluated_value
        return new_parameters

    @staticmethod
    def _process_cells(values, nb_content: NotebookNode) -> None:
        new_parameters = {}
        cells_to_process = nb_content.cells

        # TODO: if first cell has tags, insert at front instead of appending
        if cells_to_process and "metadata" in cells_to_process[0] and "parameters" in cells_to_process[0]["metadata"].get("tags", []):
            # Parse first cell for parameters
            first_cell = cells_to_process[0]
            # skip first cell
            cells_to_process = cells_to_process[1:]

            # Pull out the parameters object and ensure everything is present
            if "parameters" not in values:
                values["parameters"] = PapermillParameters()

            # Pull out the parameters object and ensure everything is present
            new_parameters = Configuration._parse_parameters_cell(first_cell)
        else:
            cells_to_process = nb_content.cells

        for cell in cells_to_process:
            cell_instance = Configuration._cell_to_content(cell)
            if cell_instance is not None:
                values["content"].middlematter.append(cell_instance)

        for k, v in new_parameters.items():
            if k in values["parameters"].model_fields and getattr(values["parameters"], k) is None:
                setattr(values["parameters"], k, v)
            elif isinstance(values["parameters"], PapermillParameters) and (k not in values["parameters"].vars or values["parameters"].vars[k] is None):
                values["parameters"].vars[k] = v

    @model_validator(mode="before")
    @classmethod
    def _append_notebook_content(cls, values) -> None:
        if values.get("notebook") is None:
            return values

        cls._init_content(values)

        file = Path(values.pop("notebook"))
        with file.open("r", encoding="utf-8") as path_file:
            nb_content = nb_read(path_file, as_version=4)

        cls._process_cells(values, nb_content)

        return values

    def generate(self, **_) -> list[NotebookNode]:
        nb = new_notebook()
        nb.metadata.nbprint = {}
        nb.metadata.nbprint.version = __version__
        nb.metadata.nbprint.tags = []
        nb.metadata.nbprint.nbprint = {}
        nb.metadata.nbprint.language = f"python{version_info.major}.{version_info.minor}"

        base_meta = {
            "tags": [],
            "nbprint": {},
        }

        nb.cells = []

        # start with parameters for papermill compat
        # use `parent=None` because we parameters is first cell, and we wont instantiate the config
        # until the next cell
        _append_or_extend(nb.cells, self.parameters.generate(metadata=base_meta.copy(), config=self, parent=None))

        # now do the configuration itself
        _append_or_extend(nb.cells, self._generate_self(metadata=base_meta.copy()))

        # now do the context object
        # pass in parent=self, attr=context so we do config.context
        _append_or_extend(nb.cells, self.context.generate(metadata=base_meta.copy(), config=self, parent=self, attr="context"))

        # resources: dict[str, SerializeAsAny[BaseModel]] = Field(default_factory=dict)
        # TODO: omitting resources, referenced directly in yaml
        # cell.metadata.nbprint.resources = {k: v.model_dump_json(by_alias=True) for k, v in self.resources.items()}

        # outputs: SerializeAsAny[Outputs]
        # TODO: skipping, consumed internally

        # now setup the page layout
        # pass in parent=self, attr=page so we do config.page
        _append_or_extend(nb.cells, self.page.generate(metadata=base_meta.copy(), config=self, parent=self, attr="page"))

        # now iterate through the content, recursively generating
        for i, content in enumerate(self.content.all):
            _append_or_extend(
                nb.cells,
                content.generate(metadata=base_meta.copy(), config=self, parent=self, attr="content", counter=i),
            )

        # Finally, run the outputs cell
        # NOTE: outputs cell doesnt usually actually do anything, unless
        # it is set to run in-context, in which case it will only
        # execute inside the notebook and note outside
        _append_or_extend(nb.cells, self.outputs.generate(metadata=base_meta.copy(), config=self, parent=self, attr="outputs"))

        return nb

    def _generate_self(self, metadata: dict) -> NotebookNode:
        cell = super()._base_generate(metadata=metadata, config=self)

        # omit the data
        cell.metadata.nbprint.data = ""

        # add extras
        cell.metadata.nbprint.debug = self.debug
        cell.metadata.nbprint.pagedjs = self.pagedjs

        # add resources
        # TODO: do this or no?
        # cell.metadata.nbprint.resources = {k: v.model_dump_json(by_alias=True) for k, v in self.resources.items()}
        cell.metadata.nbprint.outputs = self.outputs.model_dump_json(by_alias=True)
        return cell

    def _generate_resources_cells(self, metadata: dict | None = None) -> NotebookNode:
        cell = super()._base_generate(metadata=metadata, config=None)

        # omit the data
        cell.metadata.nbprint.data = ""

        # add resources
        # mod = ast.Module(body=[], type_ignores=[])
        # for k, v in self.resources.items():
        #     ...
        return cell

    @staticmethod
    def load(path_or_model: Union[str, Path, dict, "Configuration"], name: str) -> "Configuration":
        if isinstance(path_or_model, Configuration):
            return path_or_model

        if isinstance(path_or_model, str) and path_or_model.endswith(".yml"):
            raise NBPrintPathIsYamlError(path_or_model)

        if isinstance(path_or_model, str) and path_or_model.endswith(".yaml"):
            path_or_model = Path(path_or_model).resolve()

        if isinstance(path_or_model, Path):
            path_or_model = path_or_model.resolve()
            folder = str(path_or_model.parent)
            file = str(path_or_model.name)

            with initialize_config_dir(version_base=None, config_dir=folder, job_name=name):
                cfg = compose(config_name=file, overrides=[f"+name={name}"])
                config = instantiate(cfg, _convert_="all")
                if not isinstance(config, Configuration):
                    config = Configuration.model_validate(config)
                return config
        raise NBPrintPathOrModelMalformedError(path_or_model)

    def run(self, dry_run: bool = False, *, _multi: bool = False) -> Path | None:
        gen = self.generate()
        ret = None
        if self.debug:
            pprint(gen)
        if not dry_run:
            ret = self.outputs.run(self, gen)
            if ret in (None, OutputsProcessing.STOP):
                # Either a handled problem or user requested stop
                # Return None to indicate a problem
                # TODO: revisit
                return None

        if not dry_run and not self._multi and self.outputs.postprocess:
            # Run postprocessing
            self.outputs.postprocess.object([self])

            # NOTE: as of this point, we're "done"

            # reset in case we want to run again
            self._reset()
        return ret

    def _reset(self) -> None:
        # reset ourselves in case we need to rerun
        self._nb_vars = set()
        self.context._context_generated = False

    # ccflow integration
    @property
    def context_type(self) -> Type[ContextType]:
        return self.parameters.__class__

    @property
    def result_type(self) -> Type[ResultType]:
        return self.outputs.__class__

    @Flow.call
    def __call__(self, context):  # noqa: ANN204
        # NOTE: make a copy to avoid mutation during flow runs interfering with caching
        # update parameters if changed
        if context != self.parameters:
            self.parameters = context
        self.run()
        return self.outputs


load = Configuration.load



from pydantic import Field, PrivateAttr, model_validator

from nbprint.config.base import BaseModel
from nbprint.config.content import Content

__all__ = ("ContentMarshall",)


class ContentMarshall(BaseModel):
    prematter: list[Content] = Field(default_factory=list)
    frontmatter: list[Content] = Field(default_factory=list)
    middlematter: list[Content] = Field(default_factory=list)
    endmatter: list[Content] = Field(default_factory=list)

    _all: list[Content] = PrivateAttr(default_factory=list)

    @model_validator(mode="after")
    def _all(self) -> "ContentMarshall":
        self._all = self.prematter + self.frontmatter + self.middlematter + self.endmatter
        return self

    # override indexing to return from _all
    def __getitem__(self, index: int) -> Content:
        return self._all[index]

    @property
    def all(self) -> list[Content]:
        return self._all



from pathlib import Path
from typing import Literal, Tuple

from ccflow import BaseModel, PyObjectPath
from emails import Message as EmailMessage
from jinja2 import Environment
from nbformat import NotebookNode
from pydantic import Field, field_validator, model_validator

from nbprint.config import Configuration, OutputsProcessing

from .nbconvert import NBConvertOutputs

__all__ = ("SMTP", "EmailOutputs")


class SMTP(BaseModel):
    host: str = Field(..., description="SMTP server host")
    port: int | None = Field(default=25, description="SMTP server port")
    user: str | None = Field(default=None, description="SMTP server username")
    password: str | None = Field(default=None, description="SMTP server password")
    tls: bool | None = Field(default=False, description="Use TLS for SMTP connection")
    ssl: bool | None = Field(default=False, description="Use SSL for SMTP connection")
    timeout: int | None = Field(default=30, description="Timeout for SMTP connection in seconds")


class EmailOutputs(NBConvertOutputs):
    # revert this back
    target: Literal["ipynb", "html", "webhtml", "pdf", "webpdf"] | None = "ipynb"

    body: str | None = Field(default=None, description="Body of the email, defaults to output name")
    subject: str | None = Field(default=None, description="Subject of the email, defaults to output name")
    to: list[str] = Field(description="Recipient email addresses")
    from_: Tuple[str, str] | str | None = Field(default=None, description="Sender email address")
    cc: Tuple[str, str] | str | None = Field(default=None, description="CC email address")
    bcc: Tuple[str, str] | str | None = Field(default=None, description="BCC email address")

    smtp: SMTP = Field()

    postprocess: PyObjectPath = Field(
        default=PyObjectPath("nbprint.config.outputs.email.email_postprocess"),
        description=("A callable hook that is called after all processing completes to email the results."),
    )

    @model_validator(mode="after")
    def _validate_from_or_user(self) -> "EmailOutputs":
        if not self.from_ and not self.smtp.user:
            err = "Either message.from_ or smtp.user must be set"
            raise ValueError(err)
        if not self.from_:
            self.from_ = self.smtp.user
        if not self.smtp.user:
            self.smtp.user = self.from_[1] if isinstance(self.from_, tuple) else self.from_
        return self

    @field_validator("to", mode="before")
    @classmethod
    def _validate_to(cls, v) -> list[str]:
        if isinstance(v, str):
            return [v]
        return v

    @field_validator("from_")
    @classmethod
    def _validate_from(cls, v) -> Tuple[str, str] | str | None:
        # If tuple, must be (name, email)
        if isinstance(v, tuple):
            if len(v) != 2:  # noqa: PLR2004
                err = "from_ tuple must be (name, email)"
                raise ValueError(err)
            if not v[0]:
                return v[1]
        return v

    @field_validator("cc")
    @classmethod
    def _validate_cc(cls, v) -> Tuple[str, str] | str | None:
        # If tuple, must be (name, email)
        if isinstance(v, tuple):
            if len(v) != 2:  # noqa: PLR2004
                err = "cc tuple must be (name, email)"
                raise ValueError(err)
            if not v[0]:
                return v[1]
        return v

    @field_validator("bcc")
    @classmethod
    def _validate_bcc(cls, v) -> Tuple[str, str] | str | None:
        # If tuple, must be (name, email)
        if isinstance(v, tuple):
            if len(v) != 2:  # noqa: PLR2004
                err = "bcc tuple must be (name, email)"
                raise ValueError(err)
            if not v[0]:
                return v[1]
        return v

    def make_message(self, config: "Configuration") -> EmailMessage:
        env = Environment(autoescape=True)
        msg = EmailMessage(
            html=env.from_string(self.body or self.output.stem).render(config.parameters.model_dump(exclude_unset=True)),
            subject=env.from_string(self.subject or self.output.stem).render(config.parameters.model_dump(exclude_unset=True)),
            mail_from=self.from_,
            mail_to=self.to,
            cc=self.cc,
            bcc=self.bcc,
        )
        msg.attach(filename=self.output.name, content_disposition="attachment", data=self.output.read_bytes())
        return msg

    def run(self, config: "Configuration", gen: NotebookNode) -> Path:
        # generate the output file
        output_path = super().run(config=config, gen=gen)

        if output_path in (None, OutputsProcessing.STOP):
            return OutputsProcessing.STOP
        return output_path


def email_postprocess(configs: list[Configuration], comma_or_common: Literal["common", "comma"] = "common") -> None:
    """A postprocess hook to email all output files after processing completes.
    NOTE: it is assumed that all EmailOutputs instances are configured identically,
    as happens in a multi run scenario.

    Args:
        configs (list[Configuration]): The list of nbprint configuration/s.

    """
    msgs: list[EmailMessage] = []
    smtp = None

    for config in configs:
        if isinstance(config.outputs, EmailOutputs):
            msgs.append(config.outputs.make_message(config=config))
            if smtp is None:
                smtp = config.outputs.smtp.model_dump(exclude_unset=True, exclude_none=True, exclude=["type_"])
                smtp["fail_silently"] = False
    if not msgs:
        return
    # Combine message contents, adjust email and body to be the common substring, and send
    msg = msgs[0]
    if len(msgs) > 1:
        # Concatenate bodies
        combined_body = "<br>".join([m.html for m in msgs])
        msg.html = combined_body + "<br>"

        # Subject may be paramterized,
        # so find the overlapping parts of the strings
        # and replace the non-overlapping parts with "*"
        subjects = [m.subject for m in msgs]

        if comma_or_common == "comma":
            msg.subject = ", ".join(subjects)
        else:
            common_subject = subjects[0]
            for subject in subjects[1:]:
                # Find common prefix
                prefix_len = 0
                for a, b in zip(common_subject, subject, strict=False):
                    if a == b:
                        prefix_len += 1
                    else:
                        break
                # Find common suffix
                suffix_len = 0
                for a, b in zip(reversed(common_subject), reversed(subject), strict=False):
                    if a == b:
                        suffix_len += 1
                    else:
                        break
                # Build new common subject
                middle_len = max(len(common_subject), len(subject)) - prefix_len - suffix_len
                common_subject = common_subject[:prefix_len] + ("*" * middle_len)
                if suffix_len > 0:
                    common_subject += common_subject[-suffix_len:]
                break
            msg.subject = common_subject

        # Attach all attachements
        for m in msgs[1:]:
            for attachment in m.attachments:
                msg.attachments.add(attachment)

    # Send the email
    response = msg.send(to=config.outputs.to, smtp=smtp)
    if not response.success:
        err = f"Failed to send email: {response.error}"
        raise RuntimeError(err)



import sys
from datetime import date
from itertools import product
from os import environ
from os.path import exists
from unittest.mock import patch

import pytest

from nbprint import Configuration
from nbprint.cli import hydra, run


def _example_folder_does_not_exist():
    return not exists("examples")


_integration_templates = (
    "basic",
    "inline",
    "landscape",
    "finance",
    "research",
    "plotly",
    "customsize",
    "nonpagedjs",
    "greattables",
)
_integration_notebooks = (
    "basic",
    "parameters",
)
_integration_formats = (
    "ipynb",
    "html",
    "webhtml",
    "webpdf",
)


@pytest.mark.skipif(_example_folder_does_not_exist(), reason="Examples not present - skipping examples tests")
@pytest.mark.parametrize(("template", "fmt"), list(product(_integration_templates, _integration_formats)))
def test_run_e2e(template, fmt):
    if fmt == "":
        config = Configuration.load(f"examples/{template}.yaml", template)
        config.run()
    elif fmt == "webpdf":
        if template == "customsize":
            return
        run(f"examples/{template}.yaml", ["++outputs.target=webpdf"])


@pytest.mark.skipif(_example_folder_does_not_exist(), reason="Examples not present - skipping examples tests")
@pytest.mark.parametrize("parameters", ("string1", "string2"))
def test_hydra_e2e(parameters):
    run("examples/hydra.yaml", ["config=inline", "page=report", f"parameters={parameters}"])


@pytest.mark.skipif(_example_folder_does_not_exist(), reason="Examples not present - skipping examples tests")
@pytest.mark.parametrize("template", _integration_templates)
def test_email(template):
    pytest.skip("TODO", template)


@pytest.mark.skipif(_example_folder_does_not_exist(), reason="Examples - notebook direct")
@pytest.mark.parametrize(("notebook", "fmt"), list(product(_integration_notebooks, _integration_formats)))
def test_run_notebook_direct(notebook, fmt):
    if notebook == "parameters":
        run(
            f"examples/{notebook}.ipynb",
            [
                f"++nbprint.outputs.target={fmt}",
                "+nbprint.parameters.a=10",
                "+nbprint.parameters.b='hello'",
                "+nbprint.parameters.c=True",
            ],
        )
    else:
        run(f"examples/{notebook}.ipynb", [])


@pytest.mark.skipif(_example_folder_does_not_exist(), reason="Examples - notebook direct")
@pytest.mark.parametrize(("notebook", "fmt"), list(product(_integration_notebooks, _integration_formats)))
def test_email_notebook(notebook, fmt):
    if notebook == "basic":
        if not environ.get("SMTP_USER") or not environ.get("SMTP_PASSWORD"):
            pytest.skip("SMTP credentials not set in environment - skipping email test")
            return
        run(
            f"examples/{notebook}.ipynb",
            [
                "nbprint/outputs=nbprint/email",
                f"++nbprint.outputs.target={fmt}",
                f"+nbprint.outputs.to={environ['SMTP_USER']}",
                f"+nbprint.outputs.smtp.host={environ['SMTP_HOST']}",
                f"+nbprint.outputs.smtp.user={environ['SMTP_USER']}",
                f"+nbprint.outputs.smtp.password={environ['SMTP_PASSWORD']}",
                "nbprint/content/frontmatter=nbprint/title_toc",
            ],
        )
    else:
        # Skip for now
        ...


@pytest.mark.skipif(_example_folder_does_not_exist(), reason="Examples - notebook direct")
@pytest.mark.parametrize("parameters", ("string1", "string2"))
def test_run_notebook_ccflow(parameters):
    argv = [
        "nbprint",
        "--config-dir",
        "examples/ccflow",
        "nbprint=inline",
        f"+context={parameters}",
    ]
    with patch.object(sys, "argv", argv):
        hydra()


@pytest.mark.skipif(_example_folder_does_not_exist(), reason="Examples - notebook direct")
@pytest.mark.parametrize("parameters", ("string1", "string2"))
def test_run_notebook_ccflow_with_notebook(parameters):
    argv = [
        "nbprint",
        "--config-dir",
        "examples/ccflow",
        "+nbprint.notebook=examples/basic.ipynb",
        "+nbprint.name=basic",
        f"+context={parameters}",
    ]
    with patch.object(sys, "argv", argv):
        hydra()


def test_multirun():
    with patch("nbconvert.nbconvertapp.main") as mock_nbconvert_main:
        ret = run(
            "examples/basic.ipynb",
            [
                "++callable=/nbprintx",
                r"""+nbprint.outputs.naming='{{name}}-{{date}}-{{a}}'""",
                r"""+nbprintx.parameters='[{"a":1},{"a":2},{"a":3}]'""",
                "++nbprint.outputs.execute=False",
            ],
        )
        assert len(ret.outputs) == 3

        # Ensure that nbconvert was only called exactly 3 times (once per output)
        assert mock_nbconvert_main.call_count == 3


def test_email_combine_multirun():
    with patch("nbconvert.nbconvertapp.main") as mock_nbconvert_main, patch("emails.backend.smtp.backend.SMTPBackend.connection_cls") as mock_smtp_backend:
        # test inline
        # nbprint examples/basic.ipynb \
        # ++callable=/nbprintx \
        # nbprint/outputs=nbprint/email \
        # +nbprint.outputs.to=$SMTP_USER \
        # +nbprint.outputs.smtp.host=$SMTP_HOST \
        # +nbprint.outputs.smtp.user=$SMTP_USER \
        # +nbprint.outputs.smtp.password=$SMTP_PASSWORD \
        # +nbprint.outputs.target=notebook \
        # ++nbprint.outputs.execute=False \
        # +nbprintx.parameters=[{"a":1},{"a":2},{"a":3}] \
        # +nbprint.outputs.naming='\{\{name\}\}-\{\{date\}\}-\{\{a\}\}' \
        # nbprint/content/frontmatter=nbprint/title_toc

        ret = run(
            "examples/basic.ipynb",
            [
                "++callable=/nbprintx",
                "nbprint/outputs=nbprint/email",
                r"""+nbprint.outputs.naming='{{name}}-{{date}}-{{a}}'""",
                r"""+nbprintx.parameters='[{"a":1},{"a":2},{"a":3}]'""",
                "++nbprint.outputs.target=notebook",
                "++nbprint.outputs.execute=False",
                "+nbprint.outputs.to=test",
                "+nbprint.outputs.smtp.host=test",
                "+nbprint.outputs.smtp.user=test",
                "+nbprint.outputs.smtp.password=test",
                "nbprint/content/frontmatter=nbprint/title_toc",
            ],
        )
        assert len(ret.outputs) == 3

        # Ensure that nbconvert was only called exactly 3 times (once per output)
        assert mock_nbconvert_main.call_count == 3
        assert mock_smtp_backend.return_value.sendmail.call_count == 1
        msg = mock_smtp_backend.return_value.sendmail.call_args[1]
        assert msg["from_addr"] == "test"
        assert msg["to_addrs"] == ["test"]
        assert f"Subject: basic-{date.today().isoformat()}-*" in msg["msg"]



from IPython.display import Markdown
from pydantic import Field

from nbprint.config.common import FontStyle, FontWeight, HorizontalAlignment, TextDecoration, VerticalAlignment

from .base import Content


class TextComponent(Content):
    text: str | None = ""
    horizontal_alignment: HorizontalAlignment | None = None
    vertical_alignment: VerticalAlignment | None = None
    font_weight: FontWeight | None = None
    font_style: FontStyle | None = None
    text_decoration: TextDecoration | None = None

    tags: list[str] = Field(default_factory=list)

    def __call__(self, **_) -> Markdown:
        return Markdown(self.text)



from datetime import datetime
from pathlib import Path

from nbprint.cli import run


def test_outputs():
    config = run(str(Path(__file__).parent / "files" / "basic.yaml"), dry_run=True)
    assert config.outputs.naming == "{{name}}-{{date}}-{{datetime}}-{{uuid}}-{{sha}}"
    path = config.outputs.run(config=config, gen=config.generate())
    assert len(path.name) == 150
    today = datetime.now()
    assert path.stem.startswith(
        f"basic-{today.year}-{'0' + str(today.month) if today.month < 10 else today.month}-{'0' + str(today.day) if today.day < 10 else today.day}"
    )



import logging
from typing import Dict, List

from IPython.display import HTML
from pydantic import Field

from nbprint import Content

__all__ = ("LoggingBasicConfig", "LoggingConfig")


class LoggingBasicConfig(Content):
    """Set the basic logging configuration, to control which logs may show up in the report."""

    log_level: int | str = "CRITICAL"

    def __call__(self, **_) -> HTML:
        logging.basicConfig(level=self.log_level)
        return HTML("")


class LoggingConfig(Content):
    version: int = 1
    disable_existing_loggers: bool = False
    formatters: Dict[str, Dict[str, str | Dict[str, str]]] = Field(
        default={
            "simple": {"format": "[%(asctime)s][%(threadName)s][%(name)s][%(levelname)s]: %(message)s"},
            "colorlog": {
                "()": "colorlog.ColoredFormatter",
                "format": "[%(cyan)s%(asctime)s%(reset)s][%(threadName)s][%(blue)s%(name)s%(reset)s][%(log_color)s%(levelname)s%(reset)s]: %(message)s",
                "log_colors": {
                    "DEBUG": "white",
                    "INFO": "green",
                    "WARNING": "yellow",
                    "ERROR": "red",
                    "CRITICAL": "red",
                },
            },
            "whenAndWhere": {"format": "[%(asctime)s][%(threadName)s][%(name)s][%(filename)s:%(lineno)d][%(levelname)s]: %(message)s"},
        }
    )
    handlers: Dict[str, Dict[str, str]] = Field(
        default={
            "console": {
                "level": "WARNING",
                "class": "ccflow.utils.logging.StreamHandler",
                "formatter": "colorlog",
                "stream": "ext://sys.stdout",
            }
        }
    )
    root: Dict[str, str | List[str]] = Field(default={"handlers": ["console"], "level": "DEBUG"})

    def __call__(self, **_) -> HTML:
        logging.dictConfig(self.model_dump(exclude=["type_"]))
        return HTML("")



from typing import TYPE_CHECKING, Optional

from IPython.display import HTML
from nbformat import NotebookNode
from pydantic import Field, SerializeAsAny, field_validator

from nbprint.config.base import BaseModel, Role, _append_or_extend
from nbprint.config.common import Style
from nbprint.config.exceptions import NBPrintNullCellError

if TYPE_CHECKING:
    from nbprint.config.core import Configuration


class Content(BaseModel):
    content: str | list[SerializeAsAny[BaseModel]] | None = ""
    tags: list[str] = Field(default_factory=list)
    role: Role = Role.CONTENT

    # cell magics
    magics: list[str] | None = Field(default_factory=list, description="List of cell magics to apply to the cell")

    # output key
    output: str | None = Field(
        default=None,
        description="If set, the cell's outputs will be collected into the `Outputs` context under this key (only for code cells).",
    )

    # used by lots of things
    style: Style | None = None

    @field_validator("tags", mode="after")
    @classmethod
    def _ensure_tags(cls, v: list[str]) -> list[str]:
        if "nbprint:content" not in v:
            v.append("nbprint:content")
        return v

    def _postprocess_cell(self, cell) -> None:
        if isinstance(self.content, str) and self.content:
            # replace content if its a str
            cell.source = self.content

            # remove the data, redundant
            cell.metadata.nbprint.data = ""

        if self.output:
            # Add output tag
            if f"nbprint:output:{self.output}" not in cell.metadata.get("tags", []):
                cell.metadata.tags.append(f"nbprint:output:{self.output}")
            # Add to nbprint meta
            cell.metadata.nbprint.output = self.output

    def generate(
        self,
        metadata: dict | None = None,
        config: Optional["Configuration"] = None,
        parent: Optional["BaseModel"] = None,
        attr: str = "",
        counter: int | None = None,
    ) -> NotebookNode | list[NotebookNode] | None:
        # make a cell for yourself
        self_cell = super().generate(
            metadata=metadata,
            config=config,
            parent=parent,
            attr=attr or "content",
            counter=counter,
        )
        self._postprocess_cell(self_cell)
        cells = [self_cell]

        if isinstance(self.content, list):
            for i, cell in enumerate(self.content):
                _append_or_extend(
                    cells,
                    cell.generate(metadata=metadata, config=config, parent=self, attr=attr or "content", counter=i),
                )
        for cell in cells:
            if cell is None:
                raise NBPrintNullCellError

        # prefix cells with magics
        for cell in cells:
            cell: NotebookNode
            if self.magics:
                for magic in self.magics:
                    cell.source = f"%%{magic}\n{cell.source}"
        return cells

    @field_validator("content", mode="before")
    @classmethod
    def convert_content_from_obj(cls, v) -> "Content":
        if v is None:
            return []
        if isinstance(v, list):
            for i, element in enumerate(v):
                if isinstance(element, str):
                    v[i] = Content(type_=element)
                elif isinstance(element, dict):
                    v[i] = BaseModel._to_type(element)
        return v

    def __call__(self, **_) -> HTML:
        return HTML("")


class ContentMarkdown(Content):
    content: str | None = ""

    def generate(
        self,
        metadata: dict | None = None,
        config: Optional["Configuration"] = None,
        parent: Optional["BaseModel"] = None,
        **_,
    ) -> NotebookNode | list[NotebookNode] | None:
        cell = super()._base_generate_md(metadata=metadata, config=config, parent=parent)
        cell.source = self.content
        return cell


class ContentCode(Content): ...



from pathlib import Path

import pytest
from nbformat import reads

from nbprint.cli import run


@pytest.mark.parametrize("file", sorted((Path(__file__).parent / "files").glob("*")))
def test_integrations(file):
    res = run(str(file.resolve()))

    # file specific tests
    if file.name == "capture.yaml":
        output = res.outputs.notebook.read_text()
        nb = reads(output, as_version=4)
        # NOTE: last cell is outputs, so check second to last
        assert nb.cells[-2].source == '%%capture\nprint("test1")\nimport logging\nlogging.critical("test!")\n'
        assert nb.cells[-2].outputs == []

    if file.name == "plugins.yaml":
        output = res.outputs.notebook.read_text()
        assert "Example Page Content from Plugin" in output


def test_nbprint_metadata_in_basic_nb():
    c = run("examples/basic.ipynb", dry_run=True)
    assert c.content.middlematter[0].css == ":scope h2 { color: red; }"


def test_nbprint_outputs():
    # Grab config via dry run
    c = run("examples/basic.ipynb", dry_run=True)

    # Execute
    c.run()

    # Check outputs collected
    assert sorted(c.outputs.outputs.keys()) == [
        "bool",
        "file",
        "image",
        "json",
        "string",
    ]


def test_output_shortcircuit():
    # Grab config via dry run
    c = run("examples/shortcircuit.ipynb", overrides=["nbprint/outputs=nbprint/short_circuit"], dry_run=True)

    # Execute
    assert c.run() is None

    # Check outputs collected
    assert sorted(c.outputs.outputs.keys()) == ["stop"]

    # Assert that no html file called shortcircuit was created
    assert not c.outputs.output.exists()


def test_content_injection():
    res = run("examples/basic.ipynb", ["nbprint/content/frontmatter=nbprint/title_toc"])
    output = res.outputs.notebook.read_text()
    nb = reads(output, as_version=4)
    # Title cell
    assert nb.cells[4].source.startswith("# basic")
    # TOC cell
    assert nb.cells[6].source.startswith("nbprint_contenttableofcontents = nbprint_config.content[2]")


def test_parameter_injection():
    res = run("examples/parameters.ipynb", ["+nbprint.parameters.a=10", "+nbprint.parameters.b='hello'", "+nbprint.parameters.d=True"])
    output = res.outputs.notebook.read_text()
    nb = reads(output, as_version=4)
    first_cell = nb.cells[0]
    for k, v in [("a", 10), ("b", "'hello'"), ("d", True), ("c", "'abc'")]:
        assert f"{k} = {v}" in first_cell.source



from typing import TYPE_CHECKING

from nbformat import NotebookNode
from pydantic import ConfigDict, Field, PrivateAttr, field_validator

from nbprint.config.base import BaseModel, Role

from .parameters import Parameters

if TYPE_CHECKING:
    from .config import Configuration

__all__ = ("Context",)


class Context(BaseModel):
    tags: list[str] = Field(default_factory=list)
    role: Role = Role.CONTEXT
    ignore: bool = True
    parameters: Parameters | None = None

    # internals
    _nb_var_name: str = PrivateAttr(default="nbprint_ctx")
    _context_generated: bool = PrivateAttr(default=False)

    model_config = ConfigDict(
        validate_assignment=False,
        extra="allow",
        arbitrary_types_allowed=True,
    )

    @field_validator("tags", mode="after")
    @classmethod
    def _ensure_tags(cls, v: list[str]) -> list[str]:
        if "nbprint:context" not in v:
            v.append("nbprint:context")
        return v

    def generate(self, metadata: dict, config: "Configuration", parent: BaseModel, attr: str = "", **kwargs) -> NotebookNode:
        self._context_generated = True
        # TODO: may need to attach parameters manually at some point
        # in the future once papermill integration is complete
        return super().generate(metadata=metadata, config=config, parent=parent, attr=attr, **kwargs)



from pathlib import Path
from pprint import pprint

from ccflow import FlowOptions, FlowOptionsOverride, ResultBase
from ccflow.utils.hydra import cfg_explain_cli, cfg_run
from hydra import main as hydra_main
from omegaconf import OmegaConf
from typer import Argument, Option, Typer

from .config import Configuration, Executor
from .config.hydra import load_config

__all__ = ("hydra", "main", "run")


def run(
    path: str,
    overrides: list[str] | None = Argument(None),
    cfg: bool = False,
    debug: bool = False,
    dry_run: bool = False,
) -> Configuration | Executor:
    registry = load_config(path, overrides=overrides)
    model = registry["callable"] if "callable" in registry else registry["nbprint"]
    model.debug = True if debug else model.debug
    global_options = registry.get("/cli/global", FlowOptions())
    model_options = registry.get("/cli/model", FlowOptions())
    with FlowOptionsOverride(options=global_options), FlowOptionsOverride(options=model_options):
        pprint(OmegaConf.to_yaml(model.model_dump(mode="json"))) if cfg else model.run(dry_run=dry_run)
    return model


def run_cli(
    path: str,
    overrides: list[str] | None = Argument(None),
    cfg: bool = Option(False, "--cfg", is_eager=True, help="Print the config"),
    debug: bool = Option(False, "--debug", help="Run in debug mode"),
    dry_run: bool = Option(False, "--dry-run", "-d", help="Run dry run"),
) -> None:
    run(path=path, overrides=overrides, cfg=cfg, debug=debug, dry_run=dry_run)


@hydra_main(config_path=str(Path(__file__).parent / "config" / "hydra"), config_name="base", version_base=None)
def hydra(cfg) -> ResultBase:
    return cfg_run(cfg)


def hydra_explain() -> None:
    cfg_explain_cli(config_path=str(Path(__file__).parent / "config" / "hydra"), config_name="base", hydra_main=hydra)


def main() -> None:
    app = Typer()
    app.command("run")(run_cli)
    app()
